@@ -10,6 +10,11 @@
 #include "cuda/cuda_algorithm/cuda_for_each.hpp"
 #include "cuda/cuda_algorithm/cuda_transform.hpp"
 #include "cuda/cuda_algorithm/cuda_reduce.hpp"
+#include "cuda/cuda_algorithm/cuda_scan.hpp"
+#include "cuda/cuda_algorithm/cuda_merge.hpp"
+#include "cuda/cuda_algorithm/cuda_sort.hpp"
+#include "cuda/cuda_algorithm/cuda_find.hpp"
+
 
 /**
 @file cudaflow.hpp
